import logging
from typing import Union

from sqlalchemy import desc
from twisted.internet.defer import inlineCallbacks

from peek_core_device._private.server.controller.OfflineCacheController import (
    OfflineCacheController,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.GpsLocationTable import GpsLocationTable
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class DeviceInfoTableTupleProvider(TuplesProviderABC):
    def __init__(
        self,
        ormSessionCreator,
        offlineCacheController: OfflineCacheController,
        userApi,
    ):
        from peek_core_user.server.UserApiABC import UserApiABC

        self._ormSessionCreator = ormSessionCreator
        self._offlineCacheController = offlineCacheController
        self._userApi: UserApiABC = userApi

    @inlineCallbacks
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        from peek_core_user.tuples.UserLoggedInInfoTuple import (
            UserLoggedInInfoTuple,
        )

        # noinspection PyTypeChecker
        loggedInUsers: list[UserLoggedInInfoTuple] = (
            yield self._userApi.infoApi.userLoggedInInfo()
        )

        return (
            yield self._makeVortexMsgDeferred(
                filt, tupleSelector, loggedInUsers
            )
        )

    @deferToThreadWrapWithLogger(logger)
    def _makeVortexMsgDeferred(
        self, filt: dict, tupleSelector: TupleSelector, loggedInUsers
    ) -> Union[Deferred, bytes]:
        deviceId = tupleSelector.selector.get("deviceId")
        userByDeviceToken = {
            o.deviceToken: f"{o.userTitle}\n({o.userName})"
            for o in loggedInUsers
        }

        ormSession = self._ormSessionCreator()
        try:
            query = (
                ormSession.query(DeviceInfoTable, GpsLocationTable)
                .outerjoin(GpsLocationTable)
                .order_by(desc(DeviceInfoTable.lastOnline))
            )

            if deviceId is not None:
                query = query.filter(DeviceInfoTable.deviceId == deviceId)

            # Add the current location to each DeviceInfo row
            tuples = []
            for deviceInfoTableRow, gpsLocationTableRow in query.all():
                deviceInfoTableRow.currentLocation = gpsLocationTableRow
                deviceInfoTableRow.lastCacheCheck = (
                    self._offlineCacheController.lastCacheUpdate(
                        deviceInfoTableRow.deviceToken
                    ),
                )
                deviceInfoTableRow.loggedInUser = userByDeviceToken.get(
                    deviceInfoTableRow.deviceToken
                )
                tuples.append(deviceInfoTableRow)

            # Create the vortex message
            return (
                Payload(filt, tuples=tuples).makePayloadEnvelope().toVortexMsg()
            )

        finally:
            ormSession.close()



import logging
from datetime import datetime

from vortex.DeferUtil import callMethodLater
from vortex.TupleSelector import TupleSelector
from vortex.VortexUtil import debounceCall
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.DeviceUpdateTuple import (
    DeviceUpdateTuple,
)
from peek_core_device._private.tuples.OfflineCacheCombinedStatusTuple import (
    OfflineCacheCombinedStatusTuple,
)
from peek_core_device._private.tuples.OfflineCacheSettingTuple import (
    OfflineCacheSettingTuple,
)
from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple

logger = logging.getLogger(__name__)


class NotifierController:
    def __init__(self, tupleObservable: TupleDataObservableHandler):
        self._tupleObservable = tupleObservable

        from peek_core_device._private.server.DeviceApi import DeviceApi

        self._api: DeviceApi = None

    def setApi(self, api):
        self._api = api

    def shutdown(self):
        self._tupleObservable = None
        self._api = None

    @callMethodLater
    def notifyDeviceInfo(self, deviceId: str):
        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceInfoTuple.tupleName(), dict(deviceId=deviceId))
        )

        self.notifyAllDeviceInfos()

    @debounceCall(55)
    def notifyAllDeviceInfos(self):
        if not self._tupleObservable:
            return

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceInfoTuple.tupleName(), dict())
        )

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceInfoTable.tupleName(), dict())
        )

    @callMethodLater
    def notifyDeviceUpdate(self, deviceType: str):
        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                DeviceUpdateTuple.tupleName(), dict(deviceType=deviceType)
            )
        )

        self.notifyAllDeviceUpdate()

    @debounceCall(30)
    def notifyAllDeviceUpdate(self):
        if not self._tupleObservable:
            return

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceUpdateTuple.tupleName(), dict())
        )

    @callMethodLater
    def notifyDeviceOnline(self, deviceId: str, deviceToken: str, online: bool):
        """Notify Device Online

        Notify that the device has changed it's online status

        """
        self._api.notifyOfOnlineStatus(deviceId, deviceToken, online)

    @callMethodLater
    def notifyDeviceGpsLocation(
        self,
        deviceToken: str,
        latitude: float,
        longitude: float,
        updatedDate: datetime,
    ):
        self._api.notifyCurrentGpsLocation(
            deviceToken, latitude, longitude, updatedDate
        )

    @debounceCall(120)
    def notifyAllDeviceGpsLocation(self):
        if not self._tupleObservable:
            return

        from peek_core_device._private.storage.GpsLocationTable import (
            GpsLocationTable,
        )

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(GpsLocationTable.tupleName(), dict())
        )

    @callMethodLater
    def notifyDeviceOfflineCacheSetting(self, deviceToken: str):
        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                OfflineCacheSettingTuple.tupleName(),
                dict(deviceToken=deviceToken),
            )
        )

    @callMethodLater
    def notifyOfflineCacheCombinedStatusTuple(self, deviceToken: str):
        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                OfflineCacheCombinedStatusTuple.tupleName(),
                dict(deviceToken=deviceToken),
            )
        )



"""added device ip

Peek Plugin Database Migration Script

Revision ID: de24e18d0e6e
Revises: 97d44e5aeb18
Create Date: 2023-10-05 15:58:55.136116

"""

# revision identifiers, used by Alembic.
revision = "de24e18d0e6e"
down_revision = "97d44e5aeb18"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "DeviceInfo",
        sa.Column(
            "lastDeviceIp",
            sa.String(length=15),
            server_default="",
            nullable=True,
        ),
        schema="core_device",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("DeviceInfo", "lastDeviceIp", schema="core_device")
    # ### end Alembic commands ###



import logging
from collections import defaultdict
from datetime import datetime
from typing import List
from typing import Optional

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import vortexLogFailure
from vortex.Tuple import Tuple
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.tuples.OfflineCacheCombinedStatusTuple import (
    OfflineCacheCombinedStatusTuple,
)
from peek_core_device._private.tuples.OfflineCacheStatusAction import (
    OfflineCacheStatusAction,
)
from peek_core_device._private.tuples.OfflineCacheLoaderStatusTuple import (
    OfflineCacheLoaderStatusTuple,
)
from peek_core_device._private.tuples.UpdateOfflineCacheSettingAction import (
    UpdateOfflineCacheSettingAction,
)

logger = logging.getLogger(__name__)


class OfflineCacheController:
    def __init__(self, dbSessionCreator):
        self._dbSessionCreator = dbSessionCreator
        self._notifierController = None

        self._lastUpdateByDeviceToken = {}
        self._lastStatusEncodedPayloadByDeviceToken: dict[str, str] = {}

    def setNotificationController(self, notifierController: NotifierController):
        self._notifierController = notifierController

    def lastCacheUpdate(self, deviceToken: str) -> Optional[datetime]:
        return self._lastUpdateByDeviceToken.get(deviceToken)

    def lastCacheStatusEncodedPayload(self, deviceToken: str) -> str:
        return self._lastStatusEncodedPayloadByDeviceToken.get(deviceToken)

    def shutdown(self):
        self._lastUpdateByDeviceToken = {}

    def processTupleAction(self, tupleAction: TupleActionABC) -> List[Tuple]:
        if isinstance(tupleAction, UpdateOfflineCacheSettingAction):
            d = self._processOfflineCacheSettingUpdate(tupleAction)
            d.addErrback(vortexLogFailure, logger)
            return []

        if isinstance(tupleAction, OfflineCacheStatusAction):
            d = self._processOfflineCacheStatusUpdate(tupleAction)
            d.addErrback(vortexLogFailure, logger)
            return []

    @deferToThreadWrapWithLogger(logger)
    def _processOfflineCacheSettingUpdate(
        self, action: UpdateOfflineCacheSettingAction
    ) -> List[Tuple]:
        """Process Offline Cache Update

        :rtype: Deferred
        """
        ormSession = self._dbSessionCreator()
        try:
            # There should only be one item that exists if it exists.
            deviceInfo = (
                ormSession.query(DeviceInfoTable)
                .filter(DeviceInfoTable.id == action.deviceInfoId)
                .one()
            )

            # There should one be one
            deviceInfo.isOfflineCacheEnabled = action.offlineCacheEnabled

            ormSession.commit()

            self._notifierController.notifyDeviceOfflineCacheSetting(
                deviceToken=deviceInfo.deviceToken
            )

        finally:
            # Always close the session after we create it
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def _processOfflineCacheStatusUpdate(
        self, action: OfflineCacheStatusAction
    ) -> List[Tuple]:
        # lastDate = min([s.lastCheckDate for s in action.loaderStatusList])
        self._lastUpdateByDeviceToken[
            action.deviceToken
        ] = action.lastCachingStartDate

        self._lastStatusEncodedPayloadByDeviceToken[
            action.deviceToken
        ] = action.encodedCombinedTuplePayload

        self._notifierController.notifyOfflineCacheCombinedStatusTuple(
            deviceToken=action.deviceToken
        )



import logging
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_device._private.storage.Setting import FIELD_ENROLLMENT_ENABLED
from peek_core_device._private.storage.Setting import OFFICE_ENROLLMENT_ENABLED
from peek_core_device._private.storage.Setting import (
    OFFLINE_CACHE_REFRESH_SECONDS,
)
from peek_core_device._private.storage.Setting import (
    OFFLINE_MASTER_SWITCH_ENABLED,
    CHECK_BANDWIDTH_SECONDS,
    ABORT_RETRY_SECONDS,
    PAUSE_TIMEOUT_SECONDS,
    SEND_STATS_TO_SERVER_SECONDS,
)
from peek_core_device._private.storage.Setting import (
    SLOW_NETWORK_BANDWIDTH_METRIC_THRESHOLD,
)
from peek_core_device._private.storage.Setting import globalSetting
from peek_core_device._private.tuples.ClientSettingsTuple import (
    ClientSettingsTuple,
)

logger = logging.getLogger(__name__)


class ClientSettingsTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        ormSession = self._ormSessionCreator()
        try:
            clientSetting = ClientSettingsTuple()

            clientSetting.fieldEnrollmentEnabled = globalSetting(
                ormSession, FIELD_ENROLLMENT_ENABLED
            )

            clientSetting.officeEnrollmentEnabled = globalSetting(
                ormSession, OFFICE_ENROLLMENT_ENABLED
            )

            clientSetting.slowNetworkBandwidthMetricThreshold = globalSetting(
                ormSession, SLOW_NETWORK_BANDWIDTH_METRIC_THRESHOLD
            )

            clientSetting.offlineMasterSwitchEnabled = globalSetting(
                ormSession, OFFLINE_MASTER_SWITCH_ENABLED
            )

            # Timers

            clientSetting.offlineCacheSyncSeconds = globalSetting(
                ormSession, OFFLINE_CACHE_REFRESH_SECONDS
            )

            clientSetting.checkBandwidthSeconds = globalSetting(
                ormSession, CHECK_BANDWIDTH_SECONDS
            )

            clientSetting.abortRetrySeconds = globalSetting(
                ormSession, ABORT_RETRY_SECONDS
            )

            clientSetting.pauseTimeoutSeconds = globalSetting(
                ormSession, PAUSE_TIMEOUT_SECONDS
            )

            clientSetting.sendStateToServerSeconds = globalSetting(
                ormSession, SEND_STATS_TO_SERVER_SECONDS
            )

            # Create the vortex message
            return (
                Payload(filt, tuples=[clientSetting])
                .makePayloadEnvelope()
                .toVortexMsg()
            )

        finally:
            ormSession.close()



from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.PluginNames import deviceTuplePrefix
from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple


@addTupleType
class UpdateDeviceOnlineAction(TupleActionABC):
    __tupleType__ = deviceTuplePrefix + "UpdateDeviceOnlineAction"

    DEVICE_OFFLINE = DeviceInfoTuple.DEVICE_OFFLINE
    DEVICE_ONLINE = DeviceInfoTuple.DEVICE_ONLINE
    DEVICE_BACKGROUND = DeviceInfoTuple.DEVICE_BACKGROUND

    deviceId: str = TupleField()
    deviceStatus: int = TupleField()
    deviceIp: str = TupleField()



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix


@addTupleType
class WebClientVortexDetailsTuple(Tuple):
    """Web Client Vortex Details Tuple

    This tuple is sent as the first tuple on every http vortex connection from the client.

    It provides the client with the websocket vortex details
    """

    __tupleType__ = deviceTuplePrefix + "WebClientVortexDetailsTuple"

    useSsl: bool = TupleField()
    httpPort: int = TupleField()
    websocketPort: int = TupleField()
    host: str = TupleField()



"""added fileSize to update tuple

Peek Plugin Database Migration Script

Revision ID: 5dd43ff7ea8e
Revises: d94277dd8aca
Create Date: 2017-06-29 10:08:20.714750

"""

# revision identifiers, used by Alembic.
revision = "5dd43ff7ea8e"
down_revision = "d94277dd8aca"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "DeviceUpdate",
        sa.Column("fileSize", sa.Integer(), nullable=False),
        schema="core_device",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("DeviceUpdate", "fileSize", schema="core_device")
    # ### end Alembic commands ###



import logging
from collections import namedtuple
from datetime import datetime

import pytz
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall
from vortex.DeferUtil import callMethodLater
from vortex.DeferUtil import vortexLogFailure
from vortex.TupleAction import TupleActionABC
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.tuples.UpdateDeviceGpsLocationTupleAction import (
    UpdateDeviceGpsLocationTupleAction,
)
from peek_core_device.tuples.DeviceGpsLocationTuple import (
    DeviceGpsLocationTuple,
)
from peek_plugin_base.LoopingCallUtil import peekCatchErrbackWithLogger
from peek_plugin_base.storage.RunPyInPg import runPyInPg

logger = logging.getLogger(__name__)

DeviceLocationTuple = namedtuple(
    "DeviceLocationTuple",
    ["deviceToken", "latitude", "longitude", "updatedDate"],
)
TimezoneSetting = namedtuple("TimezoneSetting", ["timezone"])


class GpsController(TupleActionProcessorDelegateABC):
    INSERT_SECONDS = 120.0

    def __init__(
        self,
        dbSessionCreator,
        tupleObservable: TupleDataObservableHandler,
        notifierController: NotifierController,
    ):
        self._dbSessionCreator = dbSessionCreator
        self._tupleObservable = tupleObservable
        self._notifierController = notifierController

        self._insertQueue = []
        self._insertLoopingCall = None

    def start(self):
        self._insertLoopingCall = LoopingCall(self._poll)
        d = self._insertLoopingCall.start(self.INSERT_SECONDS)
        d.addErrback(vortexLogFailure, logger)

    def shutdown(self):
        if self._insertLoopingCall and self._insertLoopingCall.running:
            self._insertLoopingCall.stop()

        self._insertLoopingCall = None
        self._notifierController = None
        self._insertQueue = []

    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        if isinstance(tupleAction, UpdateDeviceGpsLocationTupleAction):
            self._insertQueue.append(tupleAction)
            return []

    @peekCatchErrbackWithLogger(logger)
    @inlineCallbacks
    def _poll(self):
        if not self._insertQueue:
            return

        self._insertQueue, toProcess = [], self._insertQueue

        startTime = datetime.now(pytz.UTC)
        errors = yield runPyInPg(
            logger,
            self._dbSessionCreator,
            self._insertGpsLocation,
            None,
            toProcess,
        )
        logger.debug(
            "Inserted %s GPS Locations in %s",
            len(toProcess),
            datetime.now(pytz.UTC) - startTime,
        )

        for error in errors:
            logger.error(error)

        self._notifyTuple(toProcess)

    @callMethodLater
    def _notifyTuple(self, queue: list[UpdateDeviceGpsLocationTupleAction]):
        for item in queue:
            self._tupleObservable.notifyOfTupleUpdate(
                TupleSelector(
                    DeviceGpsLocationTuple.tupleName(),
                    dict(deviceToken=item.deviceToken),
                )
            )

            self._notifierController.notifyDeviceGpsLocation(
                item.deviceToken,
                item.latitude,
                item.longitude,
                updatedDate=item.datetime,
            )

        self._notifierController.notifyAllDeviceGpsLocation()

    @classmethod
    def _insertGpsLocation(
        cls, plpy, actionTuples: list[UpdateDeviceGpsLocationTupleAction]
    ) -> list[str]:
        errors = []

        plan = plpy.prepare(
            """
            INSERT INTO core_device."GpsLocation"
             ("deviceToken", "latitude", "longitude", "updatedDate")
             VALUES
             ($1, $2, $3, $4)
             ON CONFLICT ("deviceToken")
             DO 
             UPDATE SET
                "latitude" = $2,
                "longitude" = $3,
                "updatedDate" = $4
                ;
            """,
            ["text", "float", "float", "timestamp with time zone"],
        )

        for item in list(actionTuples):
            try:
                plpy.execute(
                    plan,
                    [
                        item.deviceToken,
                        item.latitude,
                        item.longitude,
                        item.datetime,
                    ],
                )
            except plpy.SPIError as e:
                errors.append(
                    f"Failed to make update for {item.deviceToken}"
                    f": {str(e)}"
                )
                actionTuples.remove(item)

        plan = plpy.prepare(
            """
            INSERT INTO core_device."GpsLocationHistory"
             ("deviceToken", "latitude", "longitude", "loggedDate")
             VALUES
             ($1, $2, $3, $4);
            """,
            ["text", "float", "float", "timestamp with time zone"],
        )
        for item in actionTuples:
            plpy.execute(
                plan,
                [
                    item.deviceToken,
                    item.latitude,
                    item.longitude,
                    item.datetime,
                ],
            )

        return errors



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix
from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple


@addTupleType
class DeviceStatusTuple(Tuple):
    __tupleType__ = deviceTuplePrefix + "DeviceStatusTuple"

    DEVICE_OFFLINE = DeviceInfoTuple.DEVICE_OFFLINE
    DEVICE_ONLINE = DeviceInfoTuple.DEVICE_ONLINE
    DEVICE_BACKGROUND = DeviceInfoTuple.DEVICE_BACKGROUND

    deviceId: str = TupleField()
    deviceToken: str = TupleField()
    deviceStatus: int = TupleField()

    @property
    def deviceOnline(self) -> bool:
        return bool(
            self.deviceStatus & self.DEVICE_ONLINE
            and not self.deviceStatus & self.DEVICE_BACKGROUND
        )
